# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import re
import sys

from copy import copy


class Decoder():
    """
    A class to decode values of template variables.

    Template variables consist of all information stored within the YAML data of entities and between curly braces.
    Similar to variables in a programming language, template variables work as placeholders for useful information
    that must be decoded before the rendering of the output files.
    Template variables allow for the reusage of entity definitions.
    There are two types of template variables:
        - the ones defined internally by the application itself;
        - the ones defined externally by developers at the configuration file level.
    User defined template variables are set using the field "vars".

    The decoding follow an hierarchical scheme, starting with the areas that contain the robotic agents.
    These are then decoded, followed by the packages and images they temselves contain.

    Entities of type "Global Image" pass inaltered through the decoding stage.
    ...

    Attributes
    ----------
    areas : dict
        placeholder for all decoded entities of type "Area" referenced by their unique "id".

    global_images : dict
        placeholder for all decoded entities of type "Global Images" referenced by their unique "id".

    images : dict
        placeholder for all decoded entities of type "Image" referenced by their unique "id".

    packages : dict
        placeholder for all decoded entities of type "Package" referenced by their unique "id".

    robots : dict
        placeholder for all decoded entities of type "Robot" referenced by their unique "id".
    """

    areas = {}
    global_images = {}
    images = {}
    packages = {}
    robots = {}

    # The mapping between template variable names and their value.
    # This structure is the one responsible for the atual decoding of template variables.
    __decoder_variables = {}

    def __init__(self, areas, robots, images, packages, global_images):
        """
        Parameters
        ----------
        areas : dict
            placeholder for all loaded entities of type "Area" referenced by their unique "id".

        robots : dict
            placeholder for all loaded entities of type "Robot" referenced by their unique "id".

        images : dict
            placeholder for all loaded entities of type "Image" referenced by their unique "id".

        packages : dict
            placeholder for all loaded entities of type "Package" referenced by their unique "id".

        global_images : dict
            placeholder for all loaded entities of type "Global Images" referenced by their unique "id".
        """

        self.areas = areas
        self.robots = robots
        self.images = images
        self.packages = packages
        self.global_images = global_images

        # Trigger the decoding of all loaded entities.
        self.__decode()

    # Decode YAML data using the template variables mechanism
    def __substitute_template_variable(self, data):
        """Looks for and decodes all existing template variables within a sample of data.

        Besides that a verification is done to ensure that:
         - all user defined template variables were defined in the input configuration file.

        Parameters
        ----------
        data : str
            A sample of data from an entity containing template variables.

        Returns
        -------
        str
            the provided sample of data decoded.
        """

        decoded = data

        # Make sure no whitespaces exist between variable name and curly brackets.
        # TODO: remove automatically as many spaces as required. Assuming at most only one whitespace.
        decoded = decoded.replace("{{ ", "{{")
        decoded = decoded.replace(" }}", "}}")

        # Look for all template variables in between {{...}}.
        matches = re.findall("{{[A-Z_]+}}", decoded)
        for match in matches:
            try:
                # Substitute template variables by their current value.
                decoded = decoded.replace(
                    match,
                    str(self.__decoder_variables[match])
                )
            except KeyError:  # an unspecified template variable was found.
                print(f'Invalid template variable "{match}" found while decoding an instance')
                sys.exit(1)

        return decoded

    def __traverse_yaml_data(self, yaml_data):
        """Traverse the entire YAML data of an entity and substitute all existent template variables.

        Parameters
        ----------
        yaml_data : dict
            The complete YAML data of a given loaded entity.
        """

        # NOTE: The traversal of the YAML data is done recursively by the auxiliar function __traverse_yaml_data_aux.
        # This function works as a wrapper for that function simplifying the passing of arguments.
        # Therefore make sure to always call this function instead of the auxiliary one.
        self.__traverse_yaml_data_aux(yaml_data, yaml_data)

    def __traverse_yaml_data_aux(self, complete_yaml_data, yaml_data, path=''):
        """Recursive funtion to traverse the entire YAML data of an entity and substitute all existent template variables.

        This function is an auxiliary function to "__traverse_yaml_data" and should never be called directly.

        Parameters
        ----------
        complete_yaml_data : dict
            The complete YAML data of a given loaded entity.

        yaml_data : dict
            The part of YAML data of a given loaded entity that is being decoded in the current recursive step.

        path : str
            An accumulator variable (default='').
        """

        type_data = type(yaml_data)

        # Case base - a leaf node has been reached and its data must be decoded.
        # Leaf nodes are all nodes that are not of type "list" or "dict".
        if type_data in [bool, float, int, str]:
            decoded_leaf_node = self.__substitute_template_variable(str(yaml_data))  # noqa: F841
            exec(f'complete_yaml_data{path} = decoded_leaf_node')

        elif type_data in [list]:  # "list"
            for i in range(len(yaml_data)):
                self.__traverse_yaml_data_aux(complete_yaml_data, yaml_data[i], path + f"[{i}]")

        else:  # "dict"
            for field, field_value in yaml_data.items():
                # The field 'vars' contains definitions of template variables that are to be passed
                # between entities and therefore must not be decoded within the same level they are declared.
                if field not in ['vars']:
                    self.__traverse_yaml_data_aux(complete_yaml_data, field_value, path + f"['{field}']")

    def __decode_area(self, area):
        """Decode the YAML data of a single entity of type "Area".

        Parameters
        ----------
        area : object
            The entity of type "Area" to be decoded.
        """

        # Entities of type "Area" are the top of the hierarchy created between entities.
        # Therefere, whenever the YAML data of a new entity of type "Area" starts being decoded it
        # is crucial to clear all information regarding previous decodings.
        self.__decoder_variables = {}

        # Set the global tempalte variables to be passed to all other entities.
        self.__decoder_variables['{{AREA_ID}}'] = area.id

        # Trigger the decoding of the robots within the area.
        self.areas[area.id] = area
        for robot_id in area.yaml_data['robots']:
            robot = self.robots[robot_id]
            self.__decode_robot(robot)

    # flake8: noqa: C901
    def __decode_robot(self, robot):
        """Decode the YAML data of a single entity of type "Robot".

        Parameters
        ----------
        robot : object
            The entity of type "Robot" to be decoded.
        """

        # Make sure to use make a copy of the "robot" instance
        # so that the same base information may be used in decoding multiple
        # instances of type "package" and "image".
        __robot = copy(robot)

        self.__traverse_yaml_data(__robot.yaml_data)
        self.robots[__robot.id] = __robot

        # Set the global template variables to be passed to instances of type "package" and "image".
        self.__decoder_variables['{{ROBOT_ID}}'] = __robot.id
        self.__decoder_variables['{{ROBOT_ROS_DISTRO}}'] = __robot.yaml_data['ros_distro']
        if __robot.yaml_data['ros_version'] == 'ROS1':
            self.__decoder_variables['{{ROBOT_ROS_PORT}}'] = __robot.yaml_data['ros_metadata']
        else:
            self.__decoder_variables['{{ROBOT_ROS_DOMAIN}}'] = __robot.yaml_data['ros_metadata']

        # Set the custom template variables to be passed to instances of type "package" and "image".
        if 'vars' in __robot.yaml_data:
            for entry in __robot.yaml_data['vars']:
                for var, value in entry.items():
                    self.__decoder_variables['{{' + var + '}}'] = value

        # Trigger decoding of entities of type "image" associated with the robot.
        if 'images' in __robot.yaml_data:
            for image_id in __robot.yaml_data['images']:
                try:
                    image = self.images[image_id]
                    self.__decode_image(image)
                except KeyError:
                    # TODO: pass this verification step to the first stage of the pipeline.
                    print(f'Image "{image_id}" was listed but not declared.')
                    sys.exit(1)

        # Trigger decoding of entities of type "package" associated with the robot.
        if 'packages' in __robot.yaml_data:
            for package_id in __robot.yaml_data['packages']:
                try:
                    package = self.packages[package_id]
                    self.__decode_package(package)
                except KeyError:
                    # TODO: pass this verification step to the first stage of the pipeline.
                    print(f'Image "{image_id}" was listed but not declared.')
                    sys.exit(1)

    def __decode_image(self, image, is_global=False):
        """Decode the YAML data of a single entity of type "Image".

        Parameters
        ----------
        image : object
            The entity of type "Image" to be decoded.
        """

        # Make sure to use make a copy of the original "image" instance
        # so that the same base information may be used in decoding multiple instances.
        __image = copy(image)

        # Decode all template variables.
        self.__traverse_yaml_data(__image.yaml_data)

        self.images[__image.id] = __image

    def __decode_package(self, package):
        """Decode the YAML data of a single entity of type "Package".

        Parameters
        ----------
        pacakge : object
            The entity of type "Package" to be decoded.
        """

        # Make sure to use make a copy of the original "package" instance
        # so that the same base information may be used in decoding multiple instances.
        __package = copy(package)

        # Replace all template variables
        self.__traverse_yaml_data(__package.yaml_data)
        self.packages[__package.id] = __package

    def __decode(self):
        """Decode the YAML data of all loaded entities.
        """

        for _, area in self.areas.items():
            self.__decode_area(area)



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
import yaml

from .entities.area import Area
from .entities.image import Image
from .entities.package import Package
from .entities.robot import Robot


class EntityLoader():
    """
    A class used to parse an input configuration file and instantiate all entities accordingly.

    ...

    Attributes
    ----------
    areas : dict
        placeholder for all declared entities of type "Area" referenced by their unique "id".

    global_images : dict
        placeholder for all declared entities of type "Global Images" referenced by their unique "id".

    images : dict
        placeholder for all declated entities of type "Image" referenced by their unique "id".

    packages : dict
        placeholder for all declated entities of type "Package" referenced by their unique "id".

    robots : dict
        placeholder for all declated entities of type "Robot" referenced by their unique "id".

    yaml_data : dict
        the complete YAML data contained in the input configuration file.
    """

    areas = {}
    global_images = {}
    images = {}
    packages = {}
    robots = {}

    # Contains all used values of the "port" field (ROS1 robots)
    __used_ports = []

    # Contains all used values of the "domain" field (ROS2 robots)
    __used_domains = []

    def __init__(self, configuration_file_path):
        """
        Parameters
        ----------
        configuration_file_path : str
            The path for an YAML configuration input file.
        """

        self._load_data(configuration_file_path)

    # Load all entities
    # flake8: noqa: C901
    def _load_data(self, configuration_file_path):
        """Loads the contents of an YAML configuration input file.

            Parameters
            ----------
            configuration_file_path : str
                Path for the YAML configuration file to load.
            """

        # Open configuration file and ensure proper extension file.
        _, extension = os.path.splitext(configuration_file_path)
        with open(configuration_file_path, 'r') as configuration_file:
            if extension in ['.yaml', '.yml']:
                self.yaml_data = yaml.safe_load(configuration_file)
            else:
                print(f'Unsupported type "{extension}" for configuration file')
                sys.exit(1)

        # Entities are loaded hierarchically according to the following order:
        # - Areas
        #   - Robots
        #       - Packages
        #       - Images
        # - Global Images
        try:
            # Ensure that at least one entity of type "Area" is declared.
            for area in self.yaml_data['areas']:
                self._add_entity_area(area)
                try:
                    for robot_id in area['robots']:
                        # Load all declared robots within an area.
                        # NOTE: robots belonging to unknown areas will be ignored.
                        for robot in self.yaml_data['robots']:
                            if robot['id'] == robot_id:
                                self._add_entity_robot(robot, area)
                except KeyError:
                    print(f'Robot "{robot_id}" was listed but not declared.')
                    sys.exit(1)

        except KeyError:
            print('No robotic areas were declared')
            sys.exit(1)

        # Load all declared global images
        if 'globals' in self.yaml_data:
            for global_image in self.yaml_data['globals']:
                self._add_entity_image(global_image, None)

    def _add_entity_area(self, yaml_data):
        """Loads a new entity of type "Area".

        Ensures that each loaded area has an unique "id" value.

        Parameters
        ----------
        yaml_data : dict
            YAML data of the entity to be loaded.
        """

        # Load entity of type "Area".
        area = Area(yaml_data)

        # Ensures all entities have an unique "id".
        if area.id in self.areas:
            print(f'Found multiple areas with same id "{area.id}"')
            sys.exit(1)
        else:
            self.areas[area.id] = area

    # Load an entity of type 'Robot'
    # flake8: noqa: C901
    def _add_entity_robot(self, yaml_data, area):
        """Loads a new entity of type "Robot".

        For each robot loaded it triggers the loading of associated entities.
        Besides that a verification is done to ensure that:
         - each loaded robot has an unique "id" value.
         - each robot running ROS uses a different "port".
         - each robot running ROS2 uses a different "domain".

        Parameters
        ----------
        yaml_data : dict
            YAML data of the entity to be loaded.

        area : dict
            The area where the robot operates.
        """

        # Load entity of type "Robot".
        robot = Robot(yaml_data, area)

        # Ensures all entities have an unique "id".
        if robot.id in self.robots:
            print(f'Found multiple robots with same id "{robot.id}"')
            sys.exit(1)
        else:

            # Ensure unique values for "port".
            if robot.yaml_data['ros_version'] == 'ROS1':
                port = robot.yaml_data['ros_metadata']
                if port in self.__used_ports:
                    print(f'Multiple ROS robots using port "{port}"')
                    sys.exit(1)
                self.__used_ports.append(port)

            # Ensure unique values for "domain".
            if robot.yaml_data['ros_version'] == 'ROS2':
                domain = robot.yaml_data['ros_metadata']
                if domain in self.__used_domains:
                    print(f'Multiple ROS2 robots using domain "{domain}"')
                    sys.exit(1)
                self.__used_domains.append(domain)

            self.robots[robot.id] = robot

        try:
            # Load an entity of type "Package" for all packages associated with the loaded robot.
            for package in self.yaml_data['packages']:
                if f"{robot.id}-{package['id']}" in robot.yaml_data['packages']:
                    self._add_entity_package(package, robot)
        except KeyError:
            pass

        try:
            # Load an entity of type "Image" for all images associated with the loaded robot.
            for image in self.yaml_data['images']:
                if f"{robot.id}-{image['id']}" in robot.yaml_data['images']:
                    self._add_entity_image(image, robot)
        except KeyError:
            pass

    # Load an entity of type 'Image'
    def _add_entity_image(self, yaml_data, robot):
        """Loads a new entity of either type "Image" or type "Global Image".

        Besides that a verification is done to ensure that:
         - each loaded global image has an unique "id" value;
         - each loaded image has an unique "id" value.

        Parameters
        ----------
        yaml_data : dict
            YAML data of the entity to be loaded.

        robot : object
            The robotic agent where the ROS node represented by the entity will be run.
            Disambiguation between entities of type "Image" and "Global Image" is done by setting
            this parameter to None (in case an instance of the latter is to be considered).
        """

        # Load entity.
        image = Image(yaml_data, robot)

        # Select the place to store the loaded entity based on its type.
        if robot:  # storing an entity of type "Image"
            if image.id in self.images:
                print(f'Found multiple images with same id "{image.id}"')
                sys.exit(1)
            self.images[image.id] = image
        else:  # storing an entity of type "Global Image"
            if image.id in self.global_images:
                print(f'Found multiple global images with same id "{image.id}"')
                sys.exit(1)
            self.global_images[image.id] = image

    # Load an entity of type 'Package'
    def _add_entity_package(self, yaml_data, robot):
        """Loads a new entity of type "Package".

        Besides that a verification is done to ensure that:
         - each loaded package has an unique "id" value.

        Parameters
        ----------
        yaml_data : dict
            YAML data of the entity to be loaded.

        robot : object
            The robotic agent where the ROS node represented by the entity will be run.
        """

        # Load entity of type "Package".
        package = Package(yaml_data, robot)

        # Ensure that each loaded package has an unique "id".
        if package.id in self.packages:
            print(f'Found multiple packages with same id "{package.id}"')
            sys.exit(1)
        else:
            self.packages[package.id] = package



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys

# List of supported ROS and ROS2 distros.
# NOTE: probably more distros are supported - list
# only the distros that were actually tested.
ROS1_DISTROS = ['melodic', 'noetic']
ROS2_DISTROS = ['foxy']


class Robot:
    """
    A class used to represent an entity of type "Robot".
    Each entity of this type holds information about a robotic agents.

    ...

    Attributes
    ----------
    id : str
        an unique identifier for the entity.

    required_fields : list
        the required fields for an entity of this type to be considered valid.

    yaml_data : dict
        the complete YAML data for the entity obtained from the input configuration file.
    """

    # WARNING: if adding more required fields ensure that field "id" is always the first.
    # Depending on the value of required field "ros" some other fields are also required ("port" and "domain").
    # The logic that ensures these extra required fields are present is already implemented
    # in inside function "__parse_yaml_data". Therefore don't add such fields to this list.
    required_fields = ['id', 'ros']

    # flake8: noqa: C901
    def __parse_yaml_data(self, yaml_data):
        """Parses the YAML data for a given entity of type "Robot".

        In order to be considered valid, the passed YAML data must contain a non-empty entry for each required field.
        Besides that a verification is done to ensure that:
         - at least a package or an image is being declared (all existent robots must run at least one ROS node);
         - no package or image is being declared multiple times;
         - a supported ROS distribution is being selected.

        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.
        """

        # Ensure the provided YAML data contains all the required fields.
        for field in self.required_fields:

            if field not in yaml_data:
                print(f'A robot was declared without required field "{field}" .')
                sys.exit(1)

            elif field == 'id':
                # Ensure provided id is not empty
                if not yaml_data['id']:
                    print('A robot was declared with an empty "id"')
                    sys.exit(1)
                self.id = yaml_data['id']

            elif field == 'ros':

                parts = self.yaml_data['ros'].split(':')

                if not parts[0]:
                    print(f'Robot "{self.id}" was declared an empty ROS distribution.')
                    sys.exit(1)

                # The attribute "ros_version" is added to the YAML data to avoid passing
                # the arrays of distributions to later phases of the pipeline, as this
                # disambiguation between ROS and ROS2 is crucial.
                if parts[0] in ROS1_DISTROS:
                    self.yaml_data['ros_version'] = 'ROS1'
                    self.yaml_data['ros_distro'] = parts[0]
                    # Robots using ROS1 must also declare a "port".
                    if len(parts) == 1:  # use default value
                        self.yaml_data['ros_metadata'] = 11311
                    else:  # use passed value
                        if parts[-1]:
                            try:
                                self.yaml_data['ros_metadata'] = int(parts[-1])
                            except ValueError:
                                print(f'Robot "{self.id}" was declared with an invalid value for port "{parts[-1]}"')
                                sys.exit(1)
                        else:  # and invalid value for "port" was passed
                            print(f'Robot "{self.id}" was declared without a value for port')
                            sys.exit(1)

                elif parts[0] in ROS2_DISTROS:
                    self.yaml_data['ros_version'] = 'ROS2'
                    self.yaml_data['ros_distro'] = parts[0]
                    # Robots using ROS2 must also declare a "domain".
                    if len(parts) == 1:  # use default value
                        self.yaml_data['ros_metadata'] = 42
                    else:  # use passed value
                        if parts[-1]:
                            try:
                                self.yaml_data['ros_metadata'] = int(parts[-1])
                            except ValueError:
                                print(f'Robot "{self.id}" was declared with an invalid value for domain "{parts[-1]}"')
                                sys.exit(1)
                        else:
                            print(f'Robot "{self.id}" was declared without a value for domain')
                            sys.exit(1)

                # Ensure a supported ROS distribution was selected.
                else:
                    print(f'Found invalid or unsupported ROS distribution "{parts[0]}"')
                    sys.exit(1)

        environment = []
        if 'environment' in self.yaml_data:
            environment = self.yaml_data['environment']
        if yaml_data['ros_version'] == "ROS1":
            environment.append(f'ROS_HOSTNAME=roscore-{self.id}')
            environment.append(f"ROS_MASTER_URI=http://roscore-{self.id}:{yaml_data['ros_metadata']}")
        else:
            environment.append(f"ROS_DOMAIN_ID={yaml_data['ros_metadata']}")
        self.yaml_data['environment'] = environment

        # Automatically export ports so that nodes may be able to communicate with the master node.
        # This is only required for ROS1 robots.
        if yaml_data['ros_version'] == "ROS1":
            self.yaml_data['ports'] = [f"{yaml_data['ros_metadata']}:{yaml_data['ros_metadata']}"]
            self.yaml_data['command'] = f"roscore --port {yaml_data['ros_metadata']}"

        # Since a same entity of type "Package" and "Image" may be used several times
        # within distinct robots, it is crucial to ensure that each instance of a same entity has a unique "id".
        # An unique "id" is then obtained by joining the "id" of the "Robot" with the own entity's "id".

        images = []
        if 'images' in yaml_data:
            for image in yaml_data['images']:
                if not image:
                    print(f'An image with no identifier was passed within robot "{self.id}"')
                    sys.exit(1)
                # Ensure no image was declared multiple times.
                image_id = f'{self.id}-{image}'
                if image_id in images:
                    print(f'A same image "{image}" was declared multiple times within robot "{self.id}"')
                    sys.exit(1)
                images.append(image_id)
        self.yaml_data['images'] = images

        packages = []
        if 'packages' in yaml_data:
            for package in yaml_data['packages']:
                if not package:
                    print(f'A package with no identifier was passed within robot "{self.id}"')
                    sys.exit(1)
                # Ensure no package was declared multiple times.
                package_id = f'{self.id}-{package}'
                if package_id in packages:
                    print(f'A same package "{package}" was declared multiple times within robot "{self.id}"')
                    sys.exit(1)
                packages.append(package_id)
        self.yaml_data['packages'] = packages

        # Ensure that at least a package or an image is being declared.
        if not images and not packages:
            print(f'Robot "{self.id}" was declared without images or packages.')
            sys.exit(1)

        # Prepare robot networks. Even if no network is specified, all robots must be
        # automatically connected to their area's network.
        networks = []
        if 'networks' in yaml_data:
            networks = yaml_data['networks']
        networks = [f"{yaml_data['area']}-network"]
        self.yaml_data['networks'] = networks

        # The default value for "restart" is "no".
        # Unless some other value is specified set default to "always".
        if 'restart' not in yaml_data:
            self.yaml_data['restart'] = 'always'

    def __init__(self, yaml_data, area):
        """
        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.

        area : dict
            The area where the robot operates.
        """

        self.yaml_data = yaml_data
        self.yaml_data['area'] = area['id']

        self.__parse_yaml_data(yaml_data)



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys


class Image:
    """
    A class used to represent entities of both types "Image" and "Global Image".
    Each entity of type "Image" holds information about an existing Docker image that runs a ROS node and is
    associated with a robotic agent.
    Each entity of type "Global Image" holds information about an existing Docker image, that is not ROS related
    and is not associated with a particual robotic agent.

    ...

    Attributes
    ----------
    id : str
        an unique identifier for the entity.

    required_fields : list
        the required fields for an entity of this type to be considered valid.

    yaml_data : dict
        the complete YAML data for the entity obtained from the input configuration file.
    """

    # WARNING: if adding more required fields ensure that field "id" is always the first.
    required_fields = ['id', 'image']

    # flake8: noqa: C901
    def __parse_yaml_data(self, yaml_data, robot):
        """Parses the YAML data for a given entity of type "Image".

        In order to be considered valid, the passed YAML data must contain a non-empty entry for each required field.

        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.

        robot : object
            The robotic agent where the ROS node represented by the entity will be run.
            Disambiguation between entities of type "Image" and "Global Image" is done by setting
            this parameter to None (in case an instance of the latter is to be considered).
        """

        # Ensure the provided YAML data contains all the required fields
        for field in self.required_fields:

            if field not in yaml_data:
                print(f'An image was declared without required field "{field}" .')
                sys.exit(1)

            elif field == 'id':

                # Ensure provided id is not empty.
                if not yaml_data['id']:
                    print('An image was declared with an empty "id"')
                    sys.exit(1)

                # Since a same entity of type "Image" may be used several times within distinct robots,
                # it is crucial to ensure that each instance of a same entity has a unique "id".
                # An unique "id" is then obtained by joining the "id" of the "Robot" with the own entity's "id".
                # For entities of type "Global Image" this care is not required.
                if robot:
                    self.id = '{}-{}'.format(robot.id, yaml_data['id'])
                else:
                    self.id = yaml_data['id']
                self.yaml_data['id'] = self.id

            elif field == 'image':
                # Verify if Docker image's tag was specified or not.
                # If not then use default value, which depends on whether the
                # image is global or not.
                parts = yaml_data['image'].split(':')
                if len(parts) == 1:
                    if robot:
                        self.yaml_data['tag'] = '{{ROBOT_ROS_DISTRO}}'
                    else:
                        self.yaml_data['tag'] = 'latest'
                else:  # use provided tag
                    image, tag = parts
                    self.yaml_data['image'] = image
                    self.yaml_data['tag'] = tag

        # Process declared environmental variables, for entities of type "Image".
        # Add default environmental variables based on ROS version of the passed robot.
        # ROS1 distributions require the set of variables ROS_HOSTNAME and ROS_MASTER_URI.
        # ROS2 distributions require only the variable ROS_DOMAIN_ID to be set.
        environment = []
        if 'environment' in self.yaml_data:
            environment = self.yaml_data['environment']
        if robot:
            if robot.yaml_data['ros_version'] == "ROS1":
                environment.append(f'ROS_HOSTNAME={self.id}')
                environment.append('ROS_MASTER_URI=http://roscore-{{ROBOT_ID}}:{{ROBOT_ROS_PORT}}')
            else:
                environment.append('ROS_DOMAIN_ID={{ROBOT_ROS_DOMAIN}}')
        self.yaml_data['environment'] = environment

        if robot:
            # All images must also be automatically connected to the same area network as their robot.
            networks = []
            if 'networks' in yaml_data:
                networks = yaml_data['networks']
            networks.append(f"{robot.yaml_data['area']}-network")
            self.yaml_data['networks'] = networks

            # All ROS1 images must depend on the container running the master ROS node.
            depends_on = []
            if 'depends_on' in yaml_data:
                depends_on = yaml_data['depends_on']
            if robot.yaml_data['ros_version'] == 'ROS1':
                depends_on.append(f"roscore-{robot.id}")
            self.yaml_data['depends_on'] = depends_on

        # The default value for "restart" is "no".
        # Unless some other value is specified set default to "always".
        if 'restart' not in yaml_data:
            self.yaml_data['restart'] = 'always'

    def __init__(self, yaml_data, robot):
        """
        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.

        robot : object
            The robotic agent where the ROS node represented by the entity will be run.
        """
        self.yaml_data = yaml_data
        self.__parse_yaml_data(yaml_data, robot)



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Automatically containerize ROS worskpaces using Docker.
Usage:
  enlil <configuration_file> [<output_file>]
  enlil (-h | --help)
"""
import sys
from docopt import docopt
from os import path

from pipeline.decoder.decoder import Decoder
from pipeline.loader.entity_loader import EntityLoader
from pipeline.renderizer.renderizer import Renderizer


# Application entry point
def main():

    # Parse command line arguments
    arguments = docopt(__doc__)
    config_file = arguments['<configuration_file>']
    output_file = arguments['<output_file>']
    if not output_file:  # by default output file goes to current directory
        output_file = './docker-compose.yml'

    # Verify is specified configuration file exists
    if not path.isfile(config_file):
        print(f'Input configuration file "{config_file}" does not exist.')
        sys.exit(1)

    # Load all entities
    loader = EntityLoader(config_file)

    # Decode all entities
    decoder = Decoder(
        loader.areas,
        loader.robots,
        loader.images,
        loader.packages,
        loader.global_images
    )

    # Render final output file
    Renderizer(
        output_file,
        decoder.areas,
        decoder.robots,
        decoder.images,
        decoder.global_images,
        decoder.packages,
    )


if __name__ == '__main__':
    main()



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import shutil
import sys
from jinja2 import Template
from pkg_resources import resource_string


class Renderizer:
    """
    A class to render the final Docker Compose YAML file.

    Additionally a Dockerfile is also rendered for each package instance.

    Attributes
    ----------
    areas : dict
        placeholder for YAML data of entities of type "Area" referenced by their unique "id".

    direct_volumes : list
        contains a list of all the direct volumes used.

    global_images : dict
        placeholder for YAML data of entities of type "Global Images" referenced by their unique "id".

    images : dict
        placeholder for YAML data ofentities of type "Image" referenced by their unique "id".

    labeled_volumes : list
        contains a list of all the labeled volumes used.

    networks : list
        contains a list of all the networkd used.

    output_file_path : str
        the path in which to place the final Docker Compose YAML file.

    packages : dict
        placeholder for YAML data of entities of type "Package" referenced by their unique "id".

    robots : dict
        placeholder for YAML data of entities of type "Robot" referenced by their unique "id".
    """

    # initializes the data structures required for rendering
    def __init__(self, output_file_path, areas, robots, images, global_images, packages):
        """

        Parameters
        ----------
        output_file_path : str
            the path in which to place the final Docker Compose YAML file.

        areas : dict
            placeholder for all decoded entities of type "Area" referenced by their unique "id".

        robots : dict
            placeholder for all decoded entities of type "Robot" referenced by their unique "id".

        images : dict
            placeholder for all decoded entities of type "Image" referenced by their unique "id".

        global_images : dict
            placeholder for all decoded entities of type "Global Images" referenced by their unique "id".

        packages : dict
            placeholder for all decoded entities of type "Package" referenced by their unique "id".
        """
        self.output_file_path = output_file_path

        # Initialize class attributes.
        # Extract only the YAML data from the passed entities.
        self.areas = [area.yaml_data for _, area in areas.items()]
        self.global_images = [image.yaml_data for _, image in global_images.items()]
        self.images = [image.yaml_data for _, image in images.items()]
        self.packages = [package.yaml_data for _, package in packages.items()]
        self.robots = [robot.yaml_data for _, robot in robots.items()]

        # Join all volumes and networks used by all entities of type "image", "global_image" and "package"
        direct_volumes = []
        labeled_volumes = []
        networks = []

        for image in self.images:
            image_direct_volumes, image_labeled_volumes = self.__extract_volumes(image)
            direct_volumes += image_direct_volumes
            labeled_volumes += image_labeled_volumes

            networks += self.__extract_networks(image)

        for image in self.global_images:
            image_direct_volumes, image_labeled_volumes = self.__extract_volumes(image)
            direct_volumes += image_direct_volumes
            labeled_volumes += image_labeled_volumes

            networks += self.__extract_networks(image)

        for package in self.packages:
            package_direct_volumes, package_labeled_volumes = self.__extract_volumes(package)
            direct_volumes += package_direct_volumes
            labeled_volumes += package_labeled_volumes

            networks += self.__extract_networks(package)

        # Remove duplicates
        self.direct_volumes = list(dict.fromkeys(direct_volumes))
        self.labeled_volumes = list(dict.fromkeys(labeled_volumes))
        self.networks = list(dict.fromkeys(networks))

        # Render output files
        self.__render()

    def __tabs(self, number_tabs, tab_symbol='  '):
        return (f'{tab_symbol}' * number_tabs)

    def __stringify_service_robot(self, yaml_data, tabs=1):

        # ROS2 uses a distributed system and there is no master ROS node.
        if yaml_data['ros_version'] == 'ROS2':
            return ''

        final_str = f"{self.__tabs(tabs)}roscore-{yaml_data['id']}:\n"

        for field, value in yaml_data.items():

            if field not in ['area', 'images', 'packages', 'ros', 'ros_metadata', 'ros_version', 'vars']:

                type_data = type(value)

                if type_data in [bool, float, int, str]:

                    if field == 'id':
                        final_str += f"{self.__tabs(tabs + 1)}container_name: roscore-{value}\n"
                    elif field == 'ros_distro':
                        final_str += f"{self.__tabs(tabs + 1)}image: ros:{value}\n"
                    else:
                        final_str += f'{self.__tabs(tabs + 1)}{field}: {value}\n'

                elif type_data in [list]:
                    final_str += f'{self.__tabs(tabs + 1)}{field}:\n'
                    for entry in value:
                        final_str += f'{self.__tabs(tabs + 2)}- {entry}\n'

            # TODO: implement this mechanism for dictionaries
            # else:  # "dict"
            #     final_str += self.__stringify_service(field_value, tabs + f"['{field}']")

        return final_str

    def __stringify_service(self, yaml_data, tabs=1):

        final_str = f'{self.__tabs(tabs)}{yaml_data["id"]}:\n'

        for field, value in yaml_data.items():

            if field not in ['apt', 'command', 'git', 'git_cmds', 'ros', 'ssh', 'tag', 'vars']:

                type_data = type(value)

                if type_data in [bool, float, int, str]:

                    if field == 'id':
                        final_str += f"{self.__tabs(tabs + 1)}container_name: {value}\n"
                    elif field == 'image':  # an image
                        final_str += f"{self.__tabs(tabs + 1)}image: {value}:{yaml_data['tag']}\n"
                    elif field == 'path':
                        final_str += f"{self.__tabs(tabs + 1)}build: {value}{yaml_data['id']}\n"
                    else:
                        final_str += f'{self.__tabs(tabs + 1)}{field}: {value}\n'

                elif type_data in [list]:
                    if value:  # ensure the list is not empty
                        final_str += f'{self.__tabs(tabs + 1)}{field}:\n'
                        for entry in value:
                            final_str += f'{self.__tabs(tabs + 2)}- {entry}\n'

            # TODO: implement this mechanism for dictionaries
            # else:  # "dict"
            #     final_str += self.__stringify_service(field_value, tabs + f"['{field}']")

        return final_str

    def __extract_volumes(self, yaml_data):
        """Extract and filter all used volumes from an entity's YAML data.

        Parameters
        ----------
        yaml_data : dict
            An entity's YAML data.
        """

        direct_volumes = []
        labeled_volumes = []

        if 'volumes' in yaml_data:
            for volume in yaml_data['volumes']:
                parts = volume.split(':')
                if '/' in parts[0]:
                    direct_volumes.append(parts[0])
                else:  # labeled volume
                    labeled_volumes.append(parts[0])
        return direct_volumes, labeled_volumes

    def __extract_networks(self, yaml_data):
        """Extract all used networks from an entity's YAML data.

        Parameters
        ----------
        yaml_data : dict
            An entity's YAML data.
        """

        if 'networks' in yaml_data:
            return yaml_data['networks']
        return []

    def _create_folder(self, path):
        """ Create a folder to place Dockerfiles in case it does not exist already.

        Parameters
        ----------
        path : str
            Path where folder is to be created.
        """

        if not os.path.exists(path):
            os.makedirs(path)

    def __render(self):
        """  Renders the final Docker Compose YAMl file as well as all required Dockerfiles.
        """

        # Open template of Docker Compose YAML file.
        compose_yaml_template = resource_string(__name__, 'templates/docker_compose.j2').decode('utf-8')
        compose_yaml_templater = Template(compose_yaml_template)

        # Render final Docker Compose YAML file
        with open(self.output_file_path, 'w+') as output_file:
            output_file.write(compose_yaml_templater.render(
                areas=self.areas,
                direct_volumes=self.direct_volumes,
                labeled_volumes=self.labeled_volumes,
                networks=self.networks,
                stringified_global_images=[self.__stringify_service(image) for image in self.global_images],
                stringified_images=[self.__stringify_service(image) for image in self.images],
                stringified_packages=[self.__stringify_service(package) for package in self.packages],
                stringified_robots=[self.__stringify_service_robot(robot) for robot in self.robots]
            ))

        # Open template of Dockerfile.
        dockerfile_template = resource_string(__name__, 'templates/dockerfile.j2').decode('utf-8')
        dockerfile_templater = Template(dockerfile_template)

        # Create folder to place the Dockerfile for each package.
        path = '/'.join(self.output_file_path.split('/')[:-1])
        packages_path = f'{path}/packages'
        self._create_folder(packages_path)

        for package in self.packages:

            # Create a separate folder for each package.
            self._create_folder(f'{packages_path}/{package["id"]}/')

            # Create folder for SSH related files if specified.
            if 'ssh' in package:
                ssh_folder = f'{packages_path}/{package["id"]}/ssh/'
                self._create_folder(ssh_folder)
                for ssh_path in package['ssh']:
                    try:
                        file_name = ssh_path.split('/')[-1]
                        shutil.copyfile(ssh_path, ssh_folder + file_name)
                    except FileNotFoundError:
                        print(f'SSH file "{ssh_path}" was not found.')
                        sys.exit(1)

            # Render Dockerfile for each package
            with open(f'{packages_path}/{package["id"]}/Dockerfile', 'w+') as out:
                out.write(dockerfile_templater.render(package=package))



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
from distutils.util import strtobool


class Package:
    """
    A class used to represent an entity of type "Package".
    Each entity of this type holds information about a ROS node that is to be containerized
    at run-time.

    ...

    Attributes
    ----------
    id : str
        an unique identifier for the entity.

    required_fields : list
        the required fields for an entity of this type to be considered valid.

    yaml_data : dict
        the complete YAML data for the entity obtained from the input configuration file.
    """

    # WARNING: if adding more required fields ensure that field "id" is always the first.
    required_fields = ['id', 'path', 'command']

    # flake8: noqa: C901
    def __parse_yaml_data(self, yaml_data, robot):
        """Parses the YAML data for a given entity of type "Package".

        In order to be considered valid, the passed YAML data must contain a non-empty entry for each required field.

        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.

        robot : object
            The robotic agent where the ROS node represented by the entity will be run.
        """

        # Ensure the provided YAML data contains all the required fields
        for field in self.required_fields:

            if field not in yaml_data:
                print(f'A package was declared without required field "{field}" .')
                sys.exit(1)

            # Ensure provided id is not empty.
            if not yaml_data[field]:
                print(f'A package was declared with an empty "{field}"')
                sys.exit(1)

            elif field == 'id':

                # Since a same entity of type "Package" may be used several times within distinct robots,
                # it is crucial to ensure that each instance of a same entity has a unique "id".
                # An unique "id" is then obtained by joining the "id" of the "Robot" with the own entity's "id".

                self.id = '{}-{}'.format(robot.id, yaml_data['id'])
                self.yaml_data['id'] = self.id

        # If "git" field is present then add a list of the required "git clone" commands do the package YAML data.
        # This commands will later be used while rendering the package's Dockerfile, ensuring required repos are cloned.
        if 'git' in self.yaml_data:

            if not self.yaml_data['git']:
                print(f'Package "{self.id}" was declared with an empty value for "git"')
                sys.exit(1)

            git_cmds = []
            for entry in self.yaml_data['git']:
                parts = entry.split(':', 2)
                if len(parts) == 1:  # verify if "default" branch is to be used
                    git_cmds.append(f"git -C /ros_workspace/src clone -b {robot.yaml_data['ros'].split(':')[0]} {parts[0]}")
                else:
                    git_cmds.append(f"git -C /ros_workspace/src clone -b {parts[-1]} {':'.join(parts[:-1])}")
            self.yaml_data['git_cmds'] = git_cmds

        # Ensure that at least one of the fields "apt", "git" or "rosinstall" was selected.
        # There is no point in declaring an entity of this type if no ROS package is to be caontainerized.
        # NOTE: For "empty" containers entities of type "image" should be used instead.
        declared_fields = set(list(self.yaml_data.keys()))
        if not [field for field in ['apt', 'git', 'rosinstall'] if field in declared_fields]:
            print(f'Package "{self.id}" was declared without any ROS package.')
            sys.exit(1)

        # Ensure that fields "apt" and "rosinstall" are not empty, if selected.
        if 'apt' in self.yaml_data:
            if not self.yaml_data['apt']:
                print(f'Package "{self.id}" was declared with an empty value for "apt"')
                sys.exit(1)

        if 'rosinstall' in self.yaml_data:
            if not self.yaml_data['rosinstall']:
                print(f'Package "{self.id}" was declared with an empty value for "rosinstall"')
                sys.exit(1)

        # Ensure that field "ssh" is not empty 
        # Ensure that field "ssh" if a list of files
        if 'ssh' in self.yaml_data:
            if not self.yaml_data['ssh']:
                print(f'Package "{self.id}" was declared with an empty value for "ssh"')
                sys.exit(1)

            if not isinstance(self.yaml_data['ssh'], list):
                print(f'Field "ssh" of package "{self.id}" is not of type list')
                sys.exit(1)

            for filename in self.yaml_data['ssh']:
                if not isinstance(filename, str):
                    print(f'Field "ssh" of package "{self.id}" is not a list of files')
                    sys.exit(1)

        # Process declared environmental variables.
        # Add default environmental variables based on ROS version of the passed robot.
        # ROS1 distributions require the set of variables ROS_HOSTNAME and ROS_MASTER_URI.
        # ROS2 distributions require only the variable ROS_DOMAIN_ID to be set.
        environment = []
        if 'environment' in self.yaml_data:
            environment = self.yaml_data['environment']
        if robot.yaml_data['ros_version'] == "ROS1":
            environment.append(f'ROS_HOSTNAME={self.id}')
            environment.append('ROS_MASTER_URI=http://roscore-{{ROBOT_ID}}:{{ROBOT_ROS_PORT}}')
        else:
            environment.append('ROS_DOMAIN_ID={{ROBOT_ROS_DOMAIN}}')
        self.yaml_data['environment'] = environment

        # Copy the passed robot's ROS version to this instance's YAML data to later ease the
        # rendering of the package Dockerfile.
        self.yaml_data['ros'] = '{{ROBOT_ROS_DISTRO}}'

        # All packages must be automatically connected to the same area network as their robot.
        networks = []
        if 'networks' in yaml_data:
            networks = yaml_data['networks']
        networks.append(f"{robot.yaml_data['area']}-network")
        self.yaml_data['networks'] = networks

        # All ROS1 packages must depend on the container running the master ROS node.
        depends_on = []
        if 'depends_on' in yaml_data:
            depends_on = yaml_data['depends_on']
        if robot.yaml_data['ros_version'] == 'ROS1':
            depends_on.append(f"roscore-{robot.id}")
        self.yaml_data['depends_on'] = depends_on

        # The default value for "restart" is "no".
        # Unless some other value is specified set default to "always".
        if 'restart' not in yaml_data:
            self.yaml_data['restart'] = 'always'

    def __init__(self, yaml_data, robot):
        """
        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.

        robot : object
            The robotic agent where the ROS node represented by the entity will be run.
        """
        self.yaml_data = yaml_data
        self.__parse_yaml_data(yaml_data, robot)



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import setuptools

with open('./README.md', 'r') as readme:
    long_description = readme.read()

setuptools.setup(
    name='enlil',
    version='1.1.2',
    author='Pedro Melo, Rafael Arrais',
    author_email='pedro.m.melo@inesctec.pt, rafael.l.arrais@inesctec.pt',
    description='Automatic containerization of complete ROS workspaces.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/MisterOwlPT/enlil',
    license='MIT',
    packages=setuptools.find_packages(),
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.6',
    install_requires=[
        'docopt',
        'jinja2',
        'pyyaml'
    ],
    py_modules=['enlil'],
    entry_points={
        'console_scripts': [
            'enlil=enlil:main'
        ]
    },
    include_package_data=True
)



# Enlil
#
# Copyright © 2021 Pedro Pereira, Rafael Arrais
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys


class Area:
    """
    A class used to represent an entity of type "Area".
    Each entity of this type holds information about a physical area where several
    robotic agents may operate simultaneously.

    ...

    Attributes
    ----------
    id : str
        an unique identifier for the entity.

    required_fields : list
        the required fields for an entity of this type to be considered valid.

    yaml_data : dict
        the complete YAML data for the entity obtained from the input configuration file.
    """

    # WARNING: if adding more required fields ensure that field "id" is always the first.
    required_fields = ['id', 'robots']

    def __parse_yaml_data(self, yaml_data):
        """Parses the YAML data for a given entity of type "Area".

        In order to be considered valid, the passed YAML data must contain a non-empty entry for each required field.
        Besides that a verification is done to ensure that:
         - at least one robot is being declared;
         - no same robotic agent is being declared multiple times.

        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.
        """

        # Ensure the provided YAML data is not empty.
        if not len(yaml_data):
            print('An area was declared with empty data')
            sys.exit(1)

        # Ensure the provided YAML data contains all the required fields.
        for field in self.required_fields:

            if field not in yaml_data:
                print(f'An area was declared without required field "{field}"')
                sys.exit(1)

            elif field == 'id':
                # Ensure provided id is not empty.
                if not yaml_data['id']:
                    print('An area was declared with an empty "id"')
                    sys.exit(1)
                self.id = yaml_data['id']

            elif field == 'robots':
                # Ensure that a robot was not declared multiple times within a same area.
                robots = []
                for robot in yaml_data['robots']:
                    if robot in robots:
                        print(f'Robot "{robot}" was declared multiple times within area "{self.id}"')
                        sys.exit(1)
                    robots.append(robot)

                # Ensure at least one robot was declared
                if not robots:
                    print(f'Area "{self.id}" was declared without robots')
                    sys.exit(1)

    def __init__(self, yaml_data):
        """
        Parameters
        ----------
        yaml_data : dict
            The complete YAML data for the entity obtained from the input configuration file.
        """

        self.__parse_yaml_data(yaml_data)
        self.yaml_data = yaml_data
